pub mod avar;
pub mod cmap;
pub mod error;
pub mod features;
pub mod font;
pub mod fvar;
pub mod glyphs;
pub mod gvar;
pub mod head;
pub mod hvar;
pub mod metrics_and_limits;
pub mod mvar;
pub mod name;
pub mod orchestration;
pub mod os2;
pub mod paths;
pub mod post;
pub mod stat;
#[cfg(test)]
mod test_util;
